"""Utilities and helper functions."""

import importlib

from . import default_settings as settings


def import_channel(channel_alias):
    """
    helper to import channels aliases from string paths.
  
    raises an AttributeError if a channel can't be found by it's alias
    """

    try:
        channel_path = settings.NOTIFICATIONS_CHANNELS[channel_alias]
    except KeyError:
        raise AttributeError(
            '"%s" is not a valid delivery channel alias. '
            'Check your applications settings for NOTIFICATIONS_CHANNELS'
            % channel_alias
        )

    package, attr = channel_path.rsplit('.', 1)

    return getattr(importlib.import_module(package), attr)



"""Celery tasks."""

from __future__ import absolute_import, unicode_literals
from celery import shared_task

from .utils import import_channel


@shared_task
def send_notification(notification):
    """Send notification via a channel to celery."""
    for channel_alias in notification['channels']:
        channel = import_channel(channel_alias)(**notification)

        message = channel.construct_message()
        channel.notify(message)



"""Defines and listens to notification signals."""

from django.dispatch import Signal, receiver

from . import NotificationError
from .models import Notification
from .tasks import send_notification


notify = Signal(providing_args=(
    'source', 'source_display_name', 'recipient', 'action', 'category' 'obj',
    'url', 'short_description', 'extra_data', 'silent', 'channels'
))
read = Signal(providing_args=('notify_id', 'recipient'))


@receiver(notify)
def create_notification(**kwargs):
    """Notify signal receiver."""
    # make fresh copy and retain kwargs
    params = kwargs.copy()
    del params['signal']
    del params['sender']

    try:
        del params['silent']
    except KeyError:
        pass

    notification = Notification(**params)

    # If it's a not a silent notification, save the notification
    if not kwargs.get('silent', False):
        notification.save()

    # Send the notification asynchronously with celery
    send_notification.delay(notification.to_json())


@receiver(read)
def read_notification(**kwargs):
    """
    Mark notification as read.

    Raises NotificationError if the user doesn't have access
    to read the notification
    """
    notify_id = kwargs['notify_id']
    recipient = kwargs['recipient']
    notification = Notification.objects.get(id=notify_id)

    if recipient != notification.recipient:
        raise NotificationError('You cannot read this notification')

    notification.read()



"""Base Implementation of a Delivery Backend."""

import abc
from json import dumps

import six
import pika

from django.contrib.auth import get_user_model

from .models import Notification
from . import default_settings as settings


@six.add_metaclass(abc.ABCMeta)
class BaseNotificationChannel():
    """Base channel for sending notifications."""

    def __init__(self, **kwargs):
        self.notification_kwargs = kwargs

    @abc.abstractmethod
    def construct_message(self):
        """Constructs a message from notification details."""
        pass

    @abc.abstractmethod
    def notify(self, message):
        """Sends the notification."""
        pass


class ConsoleChannel(BaseNotificationChannel):
    """Dummy channel that prints to the console."""

    def construct_message(self):
        """Stringify the notification kwargs."""
        return str(self.notification_kwargs)

    def notify(self, message):
        print(message)


class BasicWebSocketChannel(BaseNotificationChannel):
    """It creates a RabbitMQ user for each user (based on their username)."""

    def _connect(self):
        """Connect to the RabbitMQ Server."""
        rabbit_mq_url = settings.NOTIFICATIONS_RABBIT_MQ_URL
        connection = pika.BlockingConnection(
            pika.connection.URLParameters(rabbit_mq_url)
        )

        return connection

    def construct_message(self):
        """Construct message from notification details."""
        
        return self.notification_kwargs

    def notify(self, message):
        """
        Puts a new message on the queue.
        
        The queue is named based on the username (for Uniqueness)
        """
        connection = self._connect()
        channel = connection.channel()

        # Get user instance
        User = get_user_model()
        source = User.objects.get(id=message['source'])
        recipient = User.objects.get(id=message['recipient'])

        channel.queue_declare(queue=source.username)

        jsonified_messasge = dumps(message)
        channel.basic_publish(
            exchange='', routing_key=recipient.username,
            body=jsonified_messasge
        )

        connection.close()



#!/usr/bin/env python

import os
from setuptools import setup


def package_files(directory):
    """Recursively add subfolders and files."""
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

EXTRA_FILES = package_files('notifications')

setup(
    name='django-notifs', version='2.6.0',
    description='Re-usable notification app for Django',
    url='https://github.com/danidee10/django-notifs', author='Osaetin Daniel',
    author_email='osaetindaniel@gmail.com', license='MIT',
    packages=['notifications'], package_data={'': EXTRA_FILES},
    install_requires=['django', 'pika', 'celery', 'six'], zip_safe=False
)



"""Register all Signals."""

from django.apps import AppConfig


class NotificationsConfig(AppConfig):
    """App Config."""
    
    name = 'notifications'

    def ready(self):
        """Import signals."""
        from . import signals
